/*
 * MacIface.cpp
 *
 *  Created on: Feb 1, 2016
 *      Author: badenhopcw
 */

#include <arpa/inet.h>
#include <iostream>
#include <cstring>
#include <assert.h>
#include "MacIface.h"
#include "PhyIface.h"
#include "MAC.h"
#include "Timer.h"

namespace Zwave {

const uint32_t MacIface::BACKOFF_WAIT[MAX_SEND_ATTEMPTS] = {40000, 80000, 160000}; // in us
const uint32_t MacIface::BEAM_WAIT[MAX_SEND_ATTEMPTS] = {3, 3, 3}; // in s (extra long to allow long beam to xmit)

MacIface::MacIface() {
	m_homeId = 0x0000;
	m_nodeId = 0x00;
	m_phy = 0;
	m_reqValidChecksum = true;
}

MacIface::~MacIface() {
	// TODO Auto-generated destructor stub
}

MacIface::MacIface (PhyIface *phy)
{
	MacIface ();
	m_phy = phy;
}

MacIface::MacIface (PhyIface *phy, uint32_t homeid, uint8_t nodeid, bool reqValidChecksum)
{
	m_phy = phy;
	m_homeId = ::htonl (homeid);
	m_nodeId = nodeid;
	m_reqValidChecksum = reqValidChecksum;
}

uint32_t
MacIface::GetHomeId_HostEnd ()
{
	return ::ntohl (m_homeId);
}

uint32_t
MacIface::GetHomeId_NetEnd ()
{
	return m_homeId;
}

uint8_t
MacIface::GetNodeId ()
{
	return m_nodeId;
}

void
MacIface::SetHomeId_NetEnd (uint32_t homeid)
{
	m_homeId = homeid;
}

void
MacIface::SetHomeId_HostEnd (uint32_t homeid)
{
	m_homeId = htonl(homeid);
}

void
MacIface::SetNodeId (uint8_t nodeid)
{
	m_nodeId = nodeid;
}
	
ssize_t
MacIface::SendFrameWaitForAck (uint8_t *frame, size_t len)
{
	
	MAC::mhr_single_t *hdr;
	FrameControl fc;
	ssize_t result;
	Timer timer;
	hdr = (MAC::mhr_single_t*)frame;
	fc = FrameControl(hdr->frameCtl);

	for (int i=0; i < MAX_SEND_ATTEMPTS; i++)
		{
			m_phy->SendFrame (frame, len, 0, 0);
			
			timer = Timer (0, BACKOFF_WAIT[i]);	
			result = RecvAckFrom (hdr->dstId, fc.GetSequenceNumber (), &timer);

			if (result != MacIface::RECVFRAME_TIMEOUT)
				return MacIface::SENDFRAME_ACK;
		}
	
	return MacIface::SENDFRAME_NOACK;
}

ssize_t
MacIface::SendBeamedFrameWaitForAck (uint8_t *frame, size_t len, uint8_t target)
{
	MAC::mhr_single_t *hdr;
	FrameControl fc;
	ssize_t result;
	Timer timer;
	hdr = (MAC::mhr_single_t*)frame;
	fc = FrameControl(hdr->frameCtl);

	for (int i=0; i < MAX_SEND_ATTEMPTS; i++)
		{
			m_phy->SendFrame (frame, len, LONGCONT_NUM_BEAMS, target); 
			//std::cout << "Setting Beam timeout to " << BEAM_WAIT[i] << " seconds.\n";
			timer = Timer (BEAM_WAIT[i],0);	
			result = RecvAckFrom (hdr->dstId, fc.GetSequenceNumber (), &timer);

			if (result != MacIface::RECVFRAME_TIMEOUT){
				std::cout << "Got ACK!\n";
				return MacIface::SENDFRAME_ACK;
			}
		}
	
	return MacIface::SENDFRAME_NOACK;

}
void
MacIface::SendFrame (uint8_t *frame, size_t len)
{
	return m_phy->SendFrame(frame, len, 0, 0);
}

ssize_t 
MacIface::Recv (uint8_t *buf, size_t maxlen, uint32_t homeid, uint8_t src, uint8_t dst, uint8_t frameType, uint8_t seqnbr, Timer *timer)
{
	MAC::mhr_fixed_t *mhr_fixed;
	MAC::mhr_single_t *mhr_single;
	size_t multi_base;
	uint8_t multi_offset;
	size_t bytesRead;
	FrameControl fc;
	size_t attempts = 0;

	while (!timer->HasTimeExpired ())
		{
			//std::cout << "MAC read attempt: " << attempts++ << std::endl;
			ssize_t phyResult = m_phy->RecvFrame(m_buf, BUFFER_SIZE, timer);
			
			if (phyResult != PhyIface::RECVFRAME_TIMEOUT)
				{
					//std::cout << "Got something...";

					assert (phyResult >=0);
					bytesRead = phyResult;

					// Always check length!
					if (bytesRead < MAC::MIN_MPDU_SIZE)
						{
							//std::cout << "too short\n";
							continue; // Too short
						}
					mhr_fixed = (Zwave::MAC::mhr_fixed_t*)m_buf;

					// Check Homeid
					if ((homeid != FILT_HOMEID_ANY) && (ntohl(mhr_fixed->homeId) != homeid))
						{
							//std::cout << "homeid doesn't match\n";						
							continue; // HomeId doesn't match
						}
					if ((src != FILT_NODEID_ANY) && (mhr_fixed->srcId != src))
						{
							//std::cout << "src ID doesn't match. Expecting " << (uint16_t)src << " but got " << (uint16_t)mhr_fixed->srcId << std::endl;
							continue; // Source ID doesn't match
						}
					if (mhr_fixed->length > bytesRead)
						{
							//std::cout << "Incorrect number of bytes available\n";
							continue; // Incorrect number of bytes available
						}
					
					fc = FrameControl (mhr_fixed->frameCtl);
					if (fc.IsMulticastFrame ())
						{
							//std::cout << "is Multicast\n";
							continue; // Don't handle multicast yet...so abort
						}
					if ((frameType != FILT_FRAMETYPE_ANY) && (fc.GetFrameType () != frameType))
						{
							//std::cout << "is wrong frame type. Expecting " << (uint16_t)frameType << " but got " << (uint16_t)fc.GetFrameType () << std::endl;
							continue; // Wrong frame
						}
					if ((seqnbr != FILT_SEQNBR_ANY) && (fc.GetSequenceNumber () != (seqnbr % FrameControl::SEQNBR_MOD)))
						{
							//std::cout << "wrong seqnbr. Expecting " << (uint16_t)(seqnbr % FrameControl::SEQNBR_MOD) << " but got " << (uint16_t)fc.GetSequenceNumber () << std::endl;						
							continue; // Wrong seqnbr
						}
					// If not multicast, then we ASSUME it is a singlecast or ACK frame, which has the same MHR
					mhr_single = (Zwave::MAC::mhr_single_t*)m_buf;
					if ((dst != FILT_NODEID_ANY) && (mhr_single->dstId != dst))
						{	
							//std::cout << "wrong destId\n";
							continue; // Incorrect destination node
						}
					if (m_reqValidChecksum)
						{
							// Make sure that length in mhr is at least bytesRead..this is assumed to have already checked above
							assert (mhr_single->length <= bytesRead); // Or overrun
							if (!MAC::CheckChecksum (m_buf, mhr_single->length))
								{
									//std::cout << "invalid checksum \n";
									continue; // Invalid checksum
								}
						}

					if (mhr_fixed->length > maxlen)
						{
							std::memcpy (buf, m_buf, maxlen);
							return MacIface::RECVFRAME_BUFOFLOW;
	 					}
					else
						{
							// We made it! This is a full frame with everything that we like in it
							std::memcpy (buf, m_buf, mhr_fixed->length); // If we have garbage on the end of our buffer, this will remove it
							return mhr_fixed->length;
						}
				}
				//std::cout << "Recv Wait Timeout\n";
		}

	return MacIface::RECVFRAME_TIMEOUT; // Bubbleup timeout condition to outside of maclayer
}

ssize_t
MacIface::RecvAny (uint8_t *buf, size_t maxlen, Timer *timer)
{
	return Recv (buf, maxlen, FILT_HOMEID_ANY, FILT_NODEID_ANY, FILT_NODEID_ANY, FILT_FRAMETYPE_ANY, FILT_SEQNBR_ANY,timer);
}

ssize_t 
MacIface::RecvNetworkAny (uint8_t *buf, size_t maxlen, Timer *timer)
{
	return Recv (buf, maxlen, GetHomeId_HostEnd (), FILT_NODEID_ANY, FILT_NODEID_ANY, FILT_FRAMETYPE_ANY, FILT_SEQNBR_ANY, timer);
}

ssize_t
MacIface::RecvNetwork (uint8_t *buf, size_t maxlen, Timer *timer)
{
	return Recv (buf, maxlen, GetHomeId_HostEnd (), FILT_NODEID_ANY, m_nodeId, FILT_FRAMETYPE_ANY, FILT_SEQNBR_ANY, timer);
}

ssize_t 
MacIface::RecvNetworkFrom (uint8_t *buf, size_t maxlen, uint8_t from, Timer *timer)
{
	return Recv (buf, maxlen, GetHomeId_HostEnd (), from, m_nodeId, FILT_FRAMETYPE_ANY, FILT_SEQNBR_ANY, timer);
}

ssize_t 
MacIface::RecvAckFrom (uint8_t from, uint8_t seqnbr, Timer *timer)
{
	MAC::mpdu_single_t frame;

	return Recv ((uint8_t*)&frame, sizeof(MAC::mpdu_single_t), GetHomeId_HostEnd (), from, m_nodeId, FrameControl::FRAME_TYPE_ACK, seqnbr, timer);
}

// CWB: Oldworld stuff below (before timeout timers)
/* 
size_t
MacIface::RecvProm (uint8_t *buf, const size_t maxlen)
{
	return m_phy->RecvFrame(buf, maxlen);
}

size_t
MacIface::RecvFromNetwork (uint8_t *buf, const size_t maxlen)
{
	// This does not handle multicast
	size_t bytesRead;
	MAC::mhr_single_t mhr;
	FrameControl fc;

	do {
		bytesRead = m_phy->RecvFrame (buf, maxlen);
		if (MAC::GetSinglecastMHR (mhr, buf, maxlen) > 0)
		{
			// The MPDU is either an ACK or Singlecast frame

			if (mhr.homeId != m_homeId) {
				std::cout << "Incorrect homeid. Expecting " << m_homeId << " but received " << mhr.homeId << std::endl;				
				bytesRead = 0;
			}

			if (mhr.dstId != m_nodeId) {
				// Need to account for bcast!
				std::cout << "Incorrect nodeId. Expecting " << (uint16_t)m_nodeId << " but received " << (uint16_t)mhr.dstId << std::endl;
				bytesRead = 0;
			}

		} else {
			// This is a multicast or malformed frame
			// TODO: Implement filtering for multicast
			std::cout << "Failed to parse MPDU header \n";			
			bytesRead = 0;
		}


	} while (bytesRead == 0);

	return bytesRead;
}

size_t
MacIface::RecvFromNetwork (uint8_t *buf, const size_t maxlen, const uint8_t src)
{
		// This does not handle multicast
	size_t bytesRead;
	MAC::mhr_single_t mhr;
	FrameControl fc;

	do {
		bytesRead = m_phy->RecvFrame (buf, maxlen);
		if (MAC::GetSinglecastMHR (mhr, buf, maxlen) > 0)
		{
			// The MPDU is either an ACK or Singlecast frame

			if (mhr.homeId != m_homeId)
				bytesRead = 0;

			if (mhr.dstId != m_nodeId) // Need to account for bcast!
				bytesRead = 0;

			if (mhr.srcId != src) // Need to account for bcast!
				bytesRead = 0;

		} else {
			// This is a multicast or malformed frame
			// TODO: Implement filtering for multicast
			bytesRead = 0;
		}


	} while (bytesRead == 0);

	return bytesRead;
}

size_t
MacIface::RecvFromNetworkProm (uint8_t *buf, const size_t maxlen, const uint8_t src, const uint8_t dst)
{
	// This does not handle multicast
	size_t bytesRead;
	MAC::mhr_single_t mhr;
	FrameControl fc;

	do {
		bytesRead = m_phy->RecvFrame (buf, maxlen);
		if (MAC::GetSinglecastMHR (mhr, buf, maxlen) > 0)
		{
			// The MPDU is either an ACK or Singlecast frame

			if (mhr.homeId != m_homeId)
				bytesRead = 0;

			if (mhr.dstId != dst) // Need to account for bcast!
				bytesRead = 0;

			if (mhr.srcId != src) // Need to account for bcast!
				bytesRead = 0;

		} else {
			// This is a multicast or malformed frame
			// TODO: Implement filtering for multicast
			bytesRead = 0;
		}


	} while (bytesRead == 0);

	return bytesRead;
*/

} /* namespace Zwave */
