#include "Security.h"
#include "SecurityHelpers.h"
#include <cryptopp/modes.h>
#include <cryptopp/aes.h>
#include <cryptopp/filters.h>
#include <cstring>
#include <string>
#include <iostream>
#include <stdint.h>

namespace Zwave {

EncPayload::EncPayload ()
{
	m_payloadLen = 0;
	m_nonceId = 0;
	std::memset (m_iv, 0, EncPayload::IV_SIZE);
	std::memset (m_payload, 0, EncPayload::MAX_ENCAP_PAYLOAD_SIZE);
	std::memset (m_cmac, 0, EncPayload::CMAC_SIZE);
}

EncPayload::~EncPayload (){}

size_t 
EncPayload::GetPackedLen ()
{
	return EncPayload::MIN_SIZE - EncPayload::MIN_ENCAP_PAYLOAD_SIZE + m_payloadLen;
}

size_t 
EncPayload::Pack(uint8_t *buf, size_t maxLen)
{
	size_t index = 0;
	if (GetPackedLen () > maxLen)
		return 0; // Not enough room in provided buffer to handle packed frame

	memcpy (&buf[index], m_iv, EncPayload::IV_SIZE);
	index += EncPayload::IV_SIZE;

	memcpy (&buf[index], m_payload, m_payloadLen);
	index += m_payloadLen;

	buf[index] = m_nonceId;
	index++;

	memcpy (&buf[index], m_cmac, EncPayload::CMAC_SIZE);
	index += EncPayload::CMAC_SIZE;

	return index;
}

size_t 
EncPayload::Unpack(uint8_t *buf, size_t maxLen)
{

	size_t index = 0;

	if (maxLen < EncPayload::MIN_SIZE)
		return 0; // Not enough room to parse the frame

  memcpy (m_iv, &buf[index], EncPayload::IV_SIZE);
	index += EncPayload::IV_SIZE;

	m_payloadLen = maxLen - (EncPayload::IV_SIZE + EncPayload::CMAC_SIZE + 1);	
	memcpy (m_payload, &buf[index], m_payloadLen);
	index += m_payloadLen;

	m_nonceId = buf[index];
	index ++;

	memcpy (m_cmac, &buf[index], EncPayload::CMAC_SIZE);
	index += EncPayload::CMAC_SIZE;

	return index; 
}

bool
Security::Decrypt (EncPayload &ep, uint8_t *encKey, uint8_t *nonce)
{

	uint8_t ivBuf[Security::BLOCK_SIZE];
	std::string decryptedTxt, encryptedTxt;
	CryptoPP::OFB_Mode< CryptoPP::AES >::Decryption d;

	// Setup full IV
	std::memcpy (ivBuf, ep.m_iv, Zwave::EncPayload::IV_SIZE);
	std::memcpy (&ivBuf[Zwave::EncPayload::IV_SIZE], nonce, Zwave::Security::NONCE_SIZE);

	try
		{
			d.SetKeyWithIV( encKey, CryptoPP::AES::DEFAULT_KEYLENGTH, ivBuf );
			encryptedTxt = std::string ((const char*)ep.m_payload, ep.m_payloadLen);

			CryptoPP::StringSource ss( encryptedTxt, true, new CryptoPP::StreamTransformationFilter( d, new CryptoPP::StringSink( decryptedTxt ))); // StringSource

			std::memcpy (ep.m_payload, (uint8_t*) decryptedTxt.c_str (), ep.m_payloadLen);
			
		}
	catch( CryptoPP::Exception& e )
		{
			std::cerr << e.what() << std::endl;
			return false;
		}

 	return true;
}

bool
Security::Encrypt (EncPayload &ep, uint8_t *encKey, uint8_t *nonce)
{
	uint8_t ivBuf[Security::BLOCK_SIZE];
	std::string plainTxt, encryptedTxt;
	CryptoPP::OFB_Mode< CryptoPP::AES >::Encryption e;

	// Setup full IV
	std::memcpy (ivBuf, ep.m_iv, Zwave::EncPayload::IV_SIZE);
	std::memcpy (&ivBuf[Zwave::EncPayload::IV_SIZE], nonce, Zwave::Security::NONCE_SIZE);

	try
		{
			e.SetKeyWithIV( encKey, CryptoPP::AES::DEFAULT_KEYLENGTH, ivBuf );
			plainTxt = std::string ((const char*)ep.m_payload, ep.m_payloadLen);

			CryptoPP::StringSource ss( plainTxt, true, new CryptoPP::StreamTransformationFilter( e, new CryptoPP::StringSink( encryptedTxt ))); // StringSource

			std::memcpy (ep.m_payload, (uint8_t*) encryptedTxt.c_str (), ep.m_payloadLen);
			
		}
	catch( CryptoPP::Exception& e )
		{
			std::cerr << e.what() << std::endl;
			return false;
		}

 	return true;
}

//CWB: I attempted to just encrypt and take the last block, but the results are not consisent with the openzwave implementation (or what is already in the calculated CMAC). This differs from Schneiers description of CBC-MAC pp99-100. 1) The IV is encrypted before the XOR with the plaintext block. 2) Open z-wave implementation only encrypts the last block. The crypto book says the blocks are encrypted after each XOR. I will have to test this to make sure I'm doing it the correct way.

// I've encrypted the IV before calling the CBC encrypt and I still don't get the correct results.
/*
bool
Security::GetCMAC2 (uint8_t *cmac, mac_message_t *macMsg, uint8_t *macKey, uint8_t *iv)
{
	uint8_t *macPtr;
	std::string plainTxt, encryptedTxt;
	uint8_t blk[Security::BLOCK_SIZE];
	
	CryptoPP::ECB_Mode< CryptoPP::AES >::Encryption eECB;
	CryptoPP::CBC_Mode< CryptoPP::AES >::Encryption eCBC;

	// Attempt to seed CBC-MAC with encrypted IV rather than the IV
	try
		{
			eECB.SetKey( macKey, KeyRing::KEY_LEN);

			// For encryption key					
			plainTxt = std::string ((const char*)iv, Security::BLOCK_SIZE);
 			encryptedTxt.clear ();
			CryptoPP::StringSource ss1( plainTxt, true, new CryptoPP::StreamTransformationFilter( eECB, new CryptoPP::StringSink(encryptedTxt ))); // StringSource
		}
	catch( CryptoPP::Exception& e )
		{
			std::cerr << e.what() << std::endl;
			return false;
		}

	PrintHex ("EncryptedIV", (uint8_t*)encryptedTxt.c_str(), Security::BLOCK_SIZE);

	plainTxt = std::string ((const char*)macMsg, macMsg->getLength ());
	
	try
		{
			eCBC.SetKeyWithIV( macKey, KeyRing::KEY_LEN, (uint8_t*)encryptedTxt.c_str());
			encryptedTxt.clear ();
			CryptoPP::StringSource ss1( plainTxt, true, new CryptoPP::StreamTransformationFilter( eCBC, new CryptoPP::StringSink(encryptedTxt ))); // StringSource
		}
	catch( CryptoPP::Exception& e )
		{
			std::cerr << e.what() << std::endl;
			return false;
		}
	
	PrintHex ("FullEncrypted", (uint8_t*)encryptedTxt.c_str (), encryptedTxt.length ());
	

	return true;
}
*/

bool
Security::GetCMAC (uint8_t *cmac, mac_message_t *macMsg, uint8_t *macKey, uint8_t *iv)
{
	uint8_t innerBlock[Security::BLOCK_SIZE];
  uint8_t *macPtr;
	std::string plainTxt, encryptedTxt;

	// CWB: Homebrewed form of CBC-MAC<AES> because documentation doesn't say its template isn't realized in dll
// and I don't want to make people rebuild the DLL
	CryptoPP::ECB_Mode< CryptoPP::AES >::Encryption e;

	// Setup the seed for the inner block (starts as the IV)
	std::memcpy(innerBlock, iv, Security::BLOCK_SIZE);

	try
		{
			e.SetKey( macKey, KeyRing::KEY_LEN);

			// For encryption key					
			plainTxt = std::string ((const char*)innerBlock, Security::BLOCK_SIZE);
 			encryptedTxt.clear ();
			CryptoPP::StringSource ss1( plainTxt, true, new CryptoPP::StreamTransformationFilter( e, new CryptoPP::StringSink(encryptedTxt ))); // StringSource
			std::memcpy (innerBlock, (uint8_t*)encryptedTxt.c_str (), KeyRing::KEY_LEN);
		}
	catch( CryptoPP::Exception& e )
		{
			std::cerr << e.what() << std::endl;
			return false;
		}

	PrintHex ("EncryptedIV", innerBlock, Security::BLOCK_SIZE);

	// Loop through all of the blocks XORing the crap out of them
	size_t msgLen = macMsg->getLength ();
	macPtr = (uint8_t*)macMsg;

	PrintHex ("MAC Msg", macPtr, msgLen);

	for (size_t i=0; i < msgLen; i++)
		{
			innerBlock[i%BLOCK_SIZE] ^= macPtr[i];
			PrintHex ("Inner Block", innerBlock, BLOCK_SIZE);

			try
				{
					// For encryption key					
					plainTxt = std::string ((const char*)innerBlock, Security::BLOCK_SIZE);
		 			encryptedTxt.clear ();
					CryptoPP::StringSource ss1( plainTxt, true, new CryptoPP::StreamTransformationFilter( e, new CryptoPP::StringSink(encryptedTxt ))); // StringSource
					std::memcpy (cmac, (uint8_t*)encryptedTxt.c_str (), EncPayload::CMAC_SIZE);  // Zwave only likes the first 8 bytes
				}
			catch( CryptoPP::Exception& e )
				{
					std::cerr << e.what() << std::endl;
					return false;
				}
		}

	//PrintHex ("XORResult", innerBlock, Security::BLOCK_SIZE);

	// Encrypt the innerBlock again for CYA props. 
	

	PrintHex ("CBC-MAC results", (uint8_t*)encryptedTxt.c_str (), Security::BLOCK_SIZE);
	return true;
}

bool
Security::GenerateCMAC (EncPayload &ep, uint8_t *macKey, uint8_t *nonce, uint8_t cmd, uint8_t src, uint8_t dst)
{
	uint8_t iv[Security::BLOCK_SIZE];
	mac_message_t macMsg;

	std::memcpy (iv, ep.m_iv, Zwave::EncPayload::IV_SIZE);
	std::memcpy (&iv[Zwave::EncPayload::IV_SIZE], nonce, Zwave::Security::NONCE_SIZE);
	GetMACMessage(macMsg, cmd, src, dst, &ep);

	return GetCMAC(ep.m_cmac, &macMsg, macKey, iv);
}

bool
Security::CheckCMAC (EncPayload *ep, uint8_t *macKey, uint8_t *nonce, uint8_t cmd, uint8_t src, uint8_t dst)
{
	uint8_t tmpCMAC[EncPayload::CMAC_SIZE];
	uint8_t iv[Security::BLOCK_SIZE];
	mac_message_t macMsg;

	GetMACMessage(macMsg, cmd, src, dst, ep);
	GetCMAC (tmpCMAC, &macMsg, macKey, iv);
	return (std::memcmp (tmpCMAC, ep->m_cmac, EncPayload::CMAC_SIZE)==0);
}


size_t 
Security::GetNonceReport (nonce_rep_t &nr, uint8_t *app, size_t app_len)
{
	size_t size = sizeof(nonce_rep_t);

	if (app_len < size)
		return 0;

	memcpy (&nr, app, size);
	return size;
}

void
Security::PrintHex (std::string header, uint8_t *buf, size_t len)
{
	size_t i;
	std::cout << header << ": " << std::hex;
	for (i=0;i<len;i++)
		{
			std::cout << " " << (uint16_t)buf[i];
		}
	std::cout << std::endl << std::dec;
	return;
}
void
Security::GetMACMessage (mac_message_t &mm, uint8_t cmd, uint8_t src, uint8_t dst, EncPayload *ep)
{
	mm.sh = cmd;
	mm.src = src;
	mm.dst = dst;
	mm.len = (uint8_t)ep->m_payloadLen;
	assert(mm.len <= mac_message_t::MAX_PAYLOAD_SIZE);

	std::memcpy(mm.payload, ep->m_payload, mm.len);
}
}
