/* Derived from 
 * ntartest.c at https://wiki.wireshark.org/Development/PcapNg
 */

#include "MacIface.h"
#include "MAC.h"
#include "APP.h"
#include "Security.h"
#include "SecurityHelpers.h"
#include "FrameControl.h"
#include "ScapyRadioIface.h"
#include <cstdio>
#include <cstring>
#include <string>
#include <iostream>
#include <assert.h>
#include <getopt.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h> /* includes net/ethernet.h */
#include <netinet/udp.h>
#include <netinet/ip.h>

#define SCAPYRADIO_RX_UDP_PORT 52002
#define MIN_BLOCK_SIZE 12
#define PCAPNG_BLOCK_TYPE_ENHANCED_PACKET 0x00000006
#define PCAPNG_BLOCK_TYPE_SIMPLE_PACKET 0x00000003

#define PCAPNG_BLOCK_HEADER_SIZE (sizeof(struct blockHeader))
#define PCAPNG_BLOCK_FOOTER_SIZE (sizeof(uint32_t))

#define MIN_PACKET_BLOCK_PAYLOAD_SIZE (sizeof(struct ether_header) + sizeof(struct iphdr) + sizeof(struct udphdr) + Zwave::ScapyRadioIface::HEADER_LENGTH + sizeof(Zwave::MAC::mhr_single_t))

struct blockHeader {
    uint32_t type;
    uint32_t length;
}__attribute__((packed));

struct enhancedPacketBlockHeader {
		uint32_t interfaceID;
		uint32_t timestampHigh;
		uint32_t timestampLow;
		uint32_t capturedLen;
		uint32_t packetLen;
}__attribute__((packed));

enum KeyFileType { ENCRYPTION, NETWORK, VERA};

//static const uint8_t encKey[Zwave::KeyRing::KEY_LEN] = {0xfc,0xcb,0x6b,0xee,0x60,0x3a,0xca,0xfd,0xf5,0xa8,0xae,0x13,0x17,0xb0,0x35,0x8e};

static uint8_t g_encKey[Zwave::KeyRing::KEY_LEN];
//static Zwave::Security::nonce_rep_t g_NonceTbl[NONCE_TABLE_SIZE];
//static size_t g_NonceIndex = 0;
static Zwave::NonceCache g_cache;
static Zwave::KeyRing g_keys;

static unsigned char data[4096];


void PrintHex (std::string header, uint8_t *buf, size_t len)
{
	size_t i;
	std::cout << header << ": " << std::hex;
	for (i=0;i<len;i++)
		{
			std::cout << " " << (uint16_t)buf[i];
		}
	std::cout << std::endl << std::dec;
	return;
}

size_t handleNonceReport (uint32_t packetNumber, uint8_t *buf, size_t bufLen)
{
	size_t result;
	size_t index;
	Zwave::Security::nonce_rep_t nonce;

	std::cout << "PacketNum=" << packetNumber << " is a nonce report frame, ";

	if (bufLen < sizeof(Zwave::Security::nonce_rep_t))
		{
			std::cout << "NOT stored due to truncated length of " << bufLen << std::endl;
			return 0;
		}	
	
	result = Zwave::Security::GetNonceReport (nonce, buf, bufLen);
	if (result > 0)
		{
			index = g_cache.StoreNonce(nonce.nonce); // fugly
			std::cout << "Stored at position " << index << " of the cache." << std::endl;
		}
	else
		{
			std::cout << "NOT stored due to parsing error." << std::endl;
		}

	std::cout << std::endl;
	return result;
}

void handleEncapEncrypt (uint32_t packetNumber, uint8_t *buf, size_t bufLen)
{

	Zwave::EncPayload enc = Zwave::EncPayload ();
	uint8_t *noncePtr = NULL; // TODO: I need to cleanup how nonce is handled. Is it a uint8_t * or nonce_rep_t?
	
	std::cout << "PacketNum=" << packetNumber << " is an encapsulated encrypted frame\n";

	enc.Unpack (buf, bufLen);
	std::cout << "nonceId=0x" << std::hex << (uint16_t)enc.m_nonceId << std::dec << " is ";
	noncePtr = g_cache.FindNonce (enc.m_nonceId);
	
	if (noncePtr != NULL)
		{
			std::cout << "FOUND in the cache." << std::endl;
			PrintHex ("Encrypted Payload", enc.m_payload, enc.m_payloadLen);
			Zwave::Security::Decrypt (enc, g_keys.m_enc, noncePtr);		
			PrintHex ("Decrypted Payload", enc.m_payload, enc.m_payloadLen);
		} 
	else
	 {
		std::cout << "NOT FOUND, so the ecapsulated frame cannot be decrypted." << std::endl;
		}

 	return;
}

void handleZwaveFrame (uint32_t packetNumber, uint8_t *buf, size_t size)
{
		size_t index = 0;
		size_t counter = 0;
		size_t frameLen = 0;
	
		Zwave::FrameControl fc;
		Zwave::MAC::mhr_single_t mhr;
		Zwave::MAC::mhr_fixed_t mhrF;
		Zwave::APP::ahr_t ahr;
		
		//std::cout << "\t\tRemaining bytes of block is " << size << std::endl;

		Zwave::MAC::GetFixedMHR (mhrF, &buf[index], size);		
		fc = Zwave::FrameControl(mhrF.frameCtl);
		frameLen = (size_t)mhrF.length;
	
		assert (frameLen <= size);
		//std::cout << "Block payload size is " << size << " and zwave frame reports a length of " << frameLen << std::endl;
		//std::cout << "\t\tlength of Zwave frame is " << frameLen << std::endl;
		// We are looking for security frames (singlecast, unrouted)
		//if (!Zwave::MAC::CheckChecksum (buf, mhrF.length)){ 
		//	std::cout << "\t\tchecksum failed\n";
		//	return;
		//}else 
		if (!fc.IsSinglecastFrame ())
			{
				//std::cout << "\t\tis not a singlecast frame\n";
				return;
			}
		else if (fc.IsRouted ())
			{
				//std::cout << "\t\tis a routed frame\n";
				return;
			}

		index += Zwave::MAC::GetSinglecastMHR (mhr, &buf[index], size);
		index += Zwave::APP::GetAppHdr (ahr, &buf[index], frameLen - index);

		//std::cout << "\t\thas cmd class of 0x" << std::hex << (uint16_t)ahr.cmd_class << std::dec << std::endl;

		// We want only security class 0x98 frames
		if (ahr.cmd_class == Zwave::Security::CMD_CLASS)
			{
				//std::cout << "\t\thas command 0x" << std::hex << (uint16_t)ahr.cmd << std::dec << std::endl;

				// APP handlers do not process checksum, so the length reduced to omit it
				if (ahr.cmd == Zwave::Security::CMD_NONCEREP)
					{
						handleNonceReport (packetNumber, &buf[index], frameLen - index - Zwave::MAC::CHECKSUM_SIZE); 
					}
				else if (ahr.cmd == Zwave::Security::CMD_ENCENCRT)
					{
						handleEncapEncrypt(packetNumber, &buf[index], frameLen - index - Zwave::MAC::CHECKSUM_SIZE);
					}
				else
					{
						//std::cout << "\t\t\tis a Security Frame, but is not a nonce report or encap frame\n"; 
					}
			}
		else
			{
				//std::cout << "\t\t\tis not a Security Frame\n";
			}
	return;
}

int loadKey (FILE *keyfile, uint8_t *keyResult)
{
	return std::fread (keyResult, 1, Zwave::KeyRing::KEY_LEN , keyfile); 
}

void genKeyFromNetworkKey (FILE *keyFile, uint8_t *keyResult)
{
	static uint8_t netKey[Zwave::KeyRing::KEY_LEN];
	loadKey(keyFile, netKey); // Load the net key

	PrintHex("NetKey", netKey, Zwave::KeyRing::KEY_LEN);
	std::memcpy (g_keys.m_net, netKey, Zwave::KeyRing::KEY_LEN);
	g_keys.GenerateFromNetworkKey ();	
	return;
}

void loadKeyFromVeraFile( FILE *keyFile, uint8_t *keyResult){
	// File is 16 x 3 bytes in length; holds three keys. 
	// We just read each key, discarding the first two 
	
	std::fseek (keyFile, Zwave::KeyRing::KEY_LEN*2, SEEK_SET); // Advance to the encrypt key
	loadKey (keyFile, keyResult); // loads encrypt key

	return;
}

void doLoadKeyFile (std::string keyPath, KeyFileType kfType)
{
		// Section loads key
		FILE *keyFile;

		keyFile = std::fopen (keyPath.c_str (), "rb");
    if (keyFile == NULL) 
			{
         std::cout << "Could not open input file " << keyPath << std::endl;
         std::exit (99);
	    }
		
		switch (kfType)
			{
				case NETWORK: genKeyFromNetworkKey (keyFile, g_keys.m_enc); break;
				case VERA: loadKeyFromVeraFile (keyFile, g_keys.m_enc); break;
				case ENCRYPTION: loadKey (keyFile, g_keys.m_enc); break;
			} 

		PrintHex ("Encryption Key", g_keys.m_enc, Zwave::KeyRing::KEY_LEN);
		std::fclose (keyFile);
		return;
}

int main(int argc, char *argv[])
{
    std::FILE *inputFile;
		std::string inputPath;
		std::string keyPath;

    int readLength;
		blockHeader blkHdr;
		//struct enhancedPacketBlockHeader *epblkHdr;
		struct ether_header *ethHdr;
		struct iphdr *ipHdr;
		struct udphdr *udpHdr;

	// Used to option flags
		bool haveKeyFile = false;
		bool haveInputFile = false;
		bool printUsage = false;
		char option_char;
		KeyFileType kfType = ENCRYPTION;

	//TODO: Do we need all of these?
    uint32_t totalRead = 0;
    int nextRead = 0;
    uint32_t blockNumber = 0;
		uint32_t packetNumber = 0;
		uint32_t blockIndex;
		uint32_t ipHdrSize;
		uint32_t blockPayloadSize;
		uint8_t *blockPayload;

		while (((option_char = getopt(argc,argv,"i:k:nvh")) != -1)&&(!printUsage))
			{
				switch (option_char){
					case 'k': haveKeyFile = true; keyPath=optarg; break;
					case 'i': haveInputFile = true; inputPath=optarg; break;
					case 'n': kfType = NETWORK; break;
					case 'v': kfType = VERA; break;
					case 'h': printUsage = true; break;
					case '?': printUsage = true; break;
				}
			}

		if (!(haveKeyFile && haveInputFile) || printUsage)
			{
				std::cout << "Usage: " << argv[0] << " -i pcapngFile -k encKeyFile [ (-n|-v) -h]\n";
				std::cout << "\tWhen -n or -v are not given, the key file is expected to be the encryption key\n";			
				std::cout << "\t-n\t Keyfile is a network key\n";
				std::cout << "\t-v\t Keyfile is in the vera key file format (found at /etc/cmh/keys)\n";
				std::cout << "\t-h\t prints this usage stuff\n\n\n";
				exit(1);
			}

		/* TODO: Consider...
			* Do we handle a compressed form?
			* Do we filter on a provided homeID, nodeID, etc?
			* Do we assume the pcap was made using scapyradio?
		  * Do we assume the interface is the lo ?
			* Do we handle old pcap files?
			* Do we handle enhanced AND simple packet blocks?
			*/

		std::cout << "InputFile: " << inputPath << std::endl;
		std::cout << "KeyFile: " << keyPath << std::endl;
		std::cout << "KeyType: ";
		switch (kfType)
			{
				case NETWORK: std::cout << "Network Key File"; break;
				case VERA: std::cout << "Vera Key File"; break;
				case ENCRYPTION: std::cout << "Encryption Key File";; break;
			} 
		std::cout << std::endl;
		doLoadKeyFile (keyPath, kfType);
		std::cout << std::endl;

		// Load input file
    inputFile = std::fopen (inputPath.c_str (), "rb");
    if(inputFile == NULL)
			{
         std::cerr << "Could not open input file " << inputPath << std::endl;
         std::exit(99);
    	}
             
		// Parse inputfile
    while(!feof(inputFile))
			{
				
				// Get block header
				readLength = std::fread(&blkHdr, 1, sizeof(blockHeader), inputFile);         
        if (readLength < sizeof(blockHeader)) 
					{
            if(readLength == 0 && feof(inputFile))
                 break;

            std::cerr << "InputFile source is too short to read a block header. Read returned with only " << readLength << " bytes.\n";
            std::exit(99);
	        }

        blockNumber++;
				
				//std::cout << "Block " << blockNumber << ";type is 0x" << std::hex << blkHdr.type << std::dec << " with size " << blkHdr.length << std::endl;


				if (blkHdr.length < MIN_BLOCK_SIZE)
					{
						std::cerr << "Encountered an invalid block size\n";
						exit(99);
					}

				// Read in the rest of the block
				readLength = std::fread (data, 1, blkHdr.length - sizeof(struct blockHeader), inputFile);
				if (readLength < (blkHdr.length - sizeof(struct blockHeader)))
					{
						std::cerr << "Block size truncated or read error. Attempted to read " << (blkHdr.length - sizeof(struct blockHeader)) << "bytes, but read returned with " << readLength << std::endl;
						std::exit(99);
					}

				// Handle only Enhanced Packet blocks
				// I may want to support simple packet block too
				if (blkHdr.type == PCAPNG_BLOCK_TYPE_ENHANCED_PACKET) 
					{
						packetNumber++;
						//std::cout << "Found enhanced block\n";

						blockPayloadSize = blkHdr.length - PCAPNG_BLOCK_HEADER_SIZE - sizeof (enhancedPacketBlockHeader) - PCAPNG_BLOCK_FOOTER_SIZE;

						if (blockPayloadSize < MIN_PACKET_BLOCK_PAYLOAD_SIZE)
							{
								//std::cerr << "Block payload size is " << blockPayloadSize << " which is smaller than " << MIN_PACKET_BLOCK_PAYLOAD_SIZE << std::endl;

								continue; // Block size is smaller than the smallest possible zwave frame so there is no point in processing the block
							}
 
						// Get the enhanced block header
						blockPayload = &data[sizeof(enhancedPacketBlockHeader)];
						blockIndex = 0;
						
						// Assume the link layer is ethernet for now, TODO: get fancier than this by checking the interface block for frame type
						// Get the ethernet header
						ethHdr = (ether_header *) &blockPayload[blockIndex];
						blockIndex += sizeof (ether_header);

				    if (ntohs (ethHdr->ether_type) == ETHERTYPE_IP) // If ether payload is IPv4
					    {
								//std::cout << "\tis IP\n";

								// Get the IP header
								ipHdr = (struct iphdr *)&blockPayload[blockIndex];
								ipHdrSize = 4 * ipHdr->ihl;

								// Need to check size again here if ipHdrSize is different than what was expected at compile time
								if (ipHdrSize > sizeof (struct iphdr))
									{
										if (blockPayloadSize < (MIN_PACKET_BLOCK_PAYLOAD_SIZE + (ipHdrSize - sizeof (struct iphdr))))
											{
												std::cerr << "Block payload size is " << blockPayloadSize << " which is smaller than " << (MIN_PACKET_BLOCK_PAYLOAD_SIZE + (ipHdrSize - sizeof (struct iphdr))) << std::endl;

												continue; // Not enough bytes to hold the smallest zwave frame, so abort further parsing
											}
									}

								blockIndex += ipHdrSize;

								if (ipHdr->protocol == IPPROTO_UDP)
									{
										//std::cout << "\tis UDP\n";
										// Get the UDP header
										udpHdr = (udphdr *)&blockPayload[blockIndex];
										blockIndex += sizeof (udphdr);

										if (ntohs(udpHdr->dest) == SCAPYRADIO_RX_UDP_PORT)
											{
												//std::cout << "\tis Scapyradio port\n";
								
												// Handle scapyradio header
												if (blockPayload[blockIndex] == Zwave::ScapyRadioIface::MAGIC_NUMBER)
													{
														//std::cout << "\thas ScapyRadio magic number\n";
														blockIndex += Zwave::ScapyRadioIface::HEADER_LENGTH;
								
														// Pass to the zwave function handler to do the rest
														handleZwaveFrame (packetNumber, &blockPayload[blockIndex], blockPayloadSize  - blockIndex);
													}
											} // scapy hdr
									}	//udp hdr					
							}	// ip hdr
					} // enhanced packet block		
			} // while (!eof())

		std::fclose(inputFile);

    return 0;
}

